/*
 * Title:        bits/list_t.hpp
 * Description:  C++ header-only wrapper for C-ABI data_t type
 * $Copyright (c) 2015-2017, 2019-2020 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: list_t.hpp 369813 2020-03-26 14:55:13Z cree $
 */

/* This file:
 * Contains the C++ implementation of list_t
 *
 */

#ifndef _DATAT_BITS_INCLUDE_TAG
#error Must only be included from sag_connectivity_cpp.hpp, use that instead
#endif

/**
 * A list class which implements many of the functions on std::vector.
 *
 * We don't guarantee exact method-for-method compatibility with any of the
 * std types.
 *
 * This class gives access methods to an ordered list of data_t.
 *
 * This is moveable, but not implicitly copyable. If you want to copy it use
 * the .copy() method. If you want to move it use std::move().
 *
 * The list is implemented as an array with amortized doubling. Modifying the
 * contents will invalidate any iterators you have and any references that you
 * have to the underlying structure.
 */
class list_t: protected sag_underlying_vector_t
{
	friend class data_t;
	friend class map_t;
	friend class Message;
	/** \internal always allocate at least a capacity of 10. */
	static size_t MIN_CAPACITY() { return 10; }
public:
	/**
	 * Forward/reverse and const/non-const iterators are implemented using this
	 * class.
	 *
	 * It implements the std::random_access_iterator concept: http://en.cppreference.com/w/cpp/concept/RandomAccessIterator
	 *
	 * You must not modify the list object while iterating over it. The iterator will throw (or assert on a non-NDEBUG build)
	 * if you try deferencing it after modifying the list object.
	 */
	template<typename DATA, typename UNDERLYING>
	struct _iterator: public std::iterator<std::random_access_iterator_tag, DATA>
	{
		typedef _iterator<DATA, UNDERLYING> iterator;
		_iterator() : list(nullptr), ptr(nullptr), direction(1) {}
		explicit _iterator(const list_t *list, UNDERLYING ptr, bool forward=true)
			: list(list), ptr(ptr), direction(forward?1:-1), count(list&&list->table?list->table->count:0)
		{}
		iterator &operator+=(size_t n) { ptr += direction*n; return *this; }
		iterator &operator-=(size_t n) { ptr -= direction*n; return *this; }
		iterator operator+(size_t n) const { return iterator(list, ptr+(direction*n), direction>0); }
		iterator operator-(size_t n) const { return iterator(list, ptr+(direction*-n), direction > 0); }
		ptrdiff_t operator-(const iterator &other) const { return direction*(ptr-other.ptr); }
		DATA &operator[](size_t n) const { return static_cast<DATA&>(ptr[n*direction]); }
		bool operator<(const iterator &other) const { return direction>0 ? ptr < other.ptr : ptr > other.ptr; }
		bool operator>(const iterator &other) const { return direction>0 ? ptr > other.ptr : ptr < other.ptr; }
		bool operator>=(const iterator &other) const { return direction>0 ? ptr >= other.ptr : ptr <= other.ptr; }
		bool operator<=(const iterator &other) const { return direction>0 ? ptr <= other.ptr : ptr >= other.ptr; }
		bool operator==(const iterator &other) const { return ptr == other.ptr; }
		bool operator!=(const iterator &other) const { return !operator==(other); }
		iterator &operator++() { return operator+=(1); }
		iterator operator++(int)
		{
			iterator tmp(*this);
			operator++();
			return tmp;
		}
		iterator &operator--() { return operator-=(1); }
		iterator operator--(int)
		{
			iterator tmp(*this);
			operator--();
			return tmp;
		}
		DATA &operator*() const {
			checkmodified();
			return static_cast<DATA&>(*ptr); 
		}
		DATA *operator->() const {
			checkmodified();
			return static_cast<DATA*>(ptr); 
		}
	protected:
		void checkmodified() const {
			if ((list && list->table && (list->table->count != count))
					|| ((!list || !list->table) && (0 != count))
				) {
				if (nullptr == getenv("AP_SKIP_DATA_T_ITERATOR_CHECKS")) {
					assert(false && "Iterator invalidated by modifying list");
					throw std::runtime_error("Iterator invalidated by modifying list");
				}
			}
		}
		const list_t *list;
		UNDERLYING ptr;
		int64_t direction;
		uint64_t count;
	};
	/** The type of iterators over a mutable list */
	typedef _iterator<data_t, sag_underlying_data_t *> iterator;
	/** The type of iterators over a const list */
	typedef _iterator<const data_t, const sag_underlying_data_t *> const_iterator;

	/** Construct an empty list */
	list_t()
	{
		table = nullptr;
	}
	/** Create a list with (at least) the specified capacity */
	explicit list_t(size_t n);

	/** Free the underlying list */
	~list_t()
	{
		clear();
	}

	/** Move assignment operator. Other will be left as the empty list and our
		previous contents freed. */
	list_t &operator=(list_t &&other)
	{
		list_t tmp;
		// put us in tmp
		swap(std::move(tmp));
		// put other in us (and by extension the null tmp in other)
		swap(std::move(other));
		return *this;
	}
	/** Move construction. Other will be left as the empty list. */
	list_t(list_t &&other)
	{
		table = nullptr;
		swap(std::move(other));
	}

	/** Initializer list construction, for example `list_t{ data_t{"item1"}, data_t{"item2"} }` */
	list_t(std::initializer_list<data_t> l)
	{
		table = nullptr;
		for(auto it = l.begin(); it != l.end(); ++it) {
			push_back(it->copy());
		}
	}

	/** Swap the contents of this list with other. */
	void swap(list_t &&other)
	{
		std::swap(table, other.table);
	}
	/** Return a deep copy of this list */
	list_t copy() const
	{
		list_t newlist(size());
		for (const_iterator it = begin(); it != end(); ++it) {
			newlist.push_back(it->copy());
		}
		return newlist;
	}

private:
	/** This is not implicitly copiable. Use the copy() method. */
	list_t &operator=(const list_t &other) { abort(); }
	/** This is not implicitly copiable. Use the copy() method. */
	list_t(const list_t &other) { abort(); }
public:

	/** Returns true if this list is deep equals to other list */
	bool operator==(const list_t &other) const;

	/** Returns false if this list is deep equals to other list */
	bool operator!=(const list_t &other) const { return !operator==(other); }

	/* Iteration: 
	 * begin() and end() methods, both const and non-const, forward and backward.
	 */
	/** Forward iterator begin */
	iterator begin() { return iterator(this, table?table->data:0); }
	/** Forward iterator end */
	iterator end() { return iterator(this, table?table->data+table->count:0); }
	/** Forward const_iterator begin */
	const_iterator begin() const { return const_iterator(this, table?table->data:0); } 
	/** Forward const_iterator end */
	const_iterator end() const { return const_iterator(this, table?table->data+table->count:0); } 
	/** Forward const_iterator begin */
	const_iterator cbegin() const { return const_iterator(this, table?table->data:0); } 
	/** Forward const_iterator end */
	const_iterator cend() const { return const_iterator(this, table?table->data+table->count:0); } 
	/** Reverse iterator begin */
	iterator rbegin() { return iterator(this, table?table->data+table->count-1:0, false); } 
	/** Reverse iterator end */
	iterator rend() { return iterator(this, table?table->data-1:0, false); } 
	/** Reverse const_iterator begin */
	const_iterator rbegin() const { return const_iterator(this, table?table->data+table->count-1:0, false); } 
	/** Reverse const_iterator end */
	const_iterator rend() const { return const_iterator(this, table?table->data-1:0, false); } 

	/** Return the number of items in this list */
	size_t size() const { return table?table->count:0; }
	/** Returns true if there are no items in this list */
	bool empty() const { return 0 == size(); }

	/** Returns a reference to the data_t at position n.
		Throws std::runtime_error if n >= size()
		The reference is only valid until this list is mutated or deleted. */
	const data_t &operator[](size_t n) const;
	/** Returns a reference to the data_t at position n.
		Throws std::runtime_error if n >= size().
		The reference is only valid until this list is mutated or deleted. */
	data_t &operator[](size_t n);
	/** Returns a reference to the data_t at position 0
		Throws std::runtime_error if the list is empty .
		The reference is only valid until this list is mutated or deleted. */
	const data_t &front() const { return operator[](0); }
	/** Returns a reference to the data_t at position size()-1
		Throws std::runtime_error if the list is empty.
		The reference is only valid until this list is mutated or deleted. */
	const data_t &back() const { return operator[](table?table->count-1:0); }
	/** Returns a reference to the data_t at position 0
		Throws std::runtime_error if the list is empty.
		The reference is only valid until this list is mutated or deleted. */
	data_t &front() { return operator[](0); }
	/** Returns a reference to the data_t at position size()-1
		Throws std::runtime_error if the list is empty.
		The reference is only valid until this list is mutated or deleted. */
	data_t &back() { return operator[](table?table->count-1:0); }

	/** Ensure the capacity is (at least) n */
	void reserve(size_t n)
	{
		if (!table) resize(n);
		if (n > table->capacity) resize(n);
	}
	/** Clear the list and free the underlying data structure
		(capacity is now 0) */
	void clear();

	/** Add an item to the end of the list.
		This is a move operation so the provided data_t will be left in an
		empty state. Use .copy() if you wish to copy into here or std::move
		otherwise. */
	void push_back(data_t &&d);

	/** Remove the last item in the list. */
	void pop_back();

private:
	/** \internal
	 *
	 * Change the capacity. It will be at least MIN_CAPACITY(). If the array
	 * currently contains items they will be copied to the new underlying
	 * array.
	 */
	void resize(size_t newsize);
};



